@@ -4,18 +4,20 @@ import json
 import os
 import sys
 from asyncio.subprocess import STDOUT
-from tempfile import NamedTemporaryFile, TemporaryDirectory
-from typing import MutableMapping, MutableSequence, Any, Tuple
+from tempfile import NamedTemporaryFile, TemporaryDirectory, mkdtemp
+from typing import MutableMapping, Any, Tuple, MutableSequence
 from typing import Optional, List
 
 import dill
+from IPython.core.compilerop import CachingCompiler
+from streamflow.core.exception import WorkflowExecutionException
 from streamflow.core.utils import get_path_processor, random_name
 from streamflow.core.workflow import Command, Status
 from streamflow.core.workflow import Job, CommandOutput, Step
 from streamflow.log_handler import logger
 from typing_extensions import Text
 
-from jupyter_workflow.ipython import executor
+from jupyter_workflow.streamflow import executor
 
 
 class JupyterCommandOutput(CommandOutput):
@@ -23,11 +25,9 @@ class JupyterCommandOutput(CommandOutput):
     def __init__(self,
                  value: Any,
                  status: Status,
-                 user_ns: MutableMapping[Text, Any],
-                 user_global_ns: Optional[MutableMapping[Text, Any]] = None):
+                 user_ns: MutableMapping[Text, Any]):
         super().__init__(value, status)
         self.user_ns: MutableMapping[Text, Any] = user_ns
-        self.user_global_ns: Optional[MutableMapping[Text, Any]] = user_global_ns
 
 
 class JupyterCommand(Command):
@@ -35,24 +35,20 @@ class JupyterCommand(Command):
     def __init__(self,
                  step: Step,
                  ast_nodes: List[Tuple[ast.AST, Text]],
-                 environment: MutableMapping[Text, Text],
+                 compiler: CachingCompiler,
                  interpreter: Text,
-                 input_names: MutableSequence[Text],
-                 output_names: MutableSequence[Text],
-                 user_ns: MutableMapping[Text, Any],
-                 user_global_ns: MutableMapping[Text, Any],
+                 input_serializers: MutableMapping[Text, Any],
+                 output_serializers: MutableMapping[Text, Any],
                  autoawait: bool,
                  stderr: Optional[Text] = None,
                  stdin: Optional[Text] = None,
                  stdout: Optional[Text] = None):
         super().__init__(step)
         self.ast_nodes: List[Tuple[ast.AST, Text]] = ast_nodes
-        self.environment: MutableMapping[Text, Text] = environment
+        self.compiler: CachingCompiler = compiler
         self.interpreter: Text = interpreter
-        self.input_names: MutableSequence[Text] = input_names
-        self.output_names: MutableSequence[Text] = output_names
-        self.user_ns: MutableMapping[Text, Any] = user_ns
-        self.user_global_ns: MutableMapping[Text, Any] = user_global_ns
+        self.input_serializers: MutableMapping[Text, Any] = input_serializers
+        self.output_serializers: MutableMapping[Text, Any] = output_serializers
         self.autoawait: bool = autoawait
         self.stderr: Optional[Text] = stderr
         self.stdin: Optional[Text] = stdin
@@ -69,10 +65,41 @@ class JupyterCommand(Command):
                     dst=dest_path,
                     dst_job=None)
                 with open(dest_path, 'rb') as f:
-                    return dill.load(f)
+                    namespace = dill.load(f)
+                for name, value in namespace.items():
+                    if name in self.output_serializers:
+                        intermediate_type = self.output_serializers[name].get('type', 'name')
+                        if intermediate_type == 'file':
+                            dest_path = os.path.join(mkdtemp(), path_processor.basename(namespace[name]))
+                            await self.step.context.data_manager.transfer_data(
+                                src=namespace[name],
+                                src_job=job,
+                                dst=dest_path,
+                                dst_job=None)
+                            namespace[name] = dest_path
+                return {k: executor.postload(
+                    compiler=self.compiler,
+                    name=k,
+                    value=v,
+                    serializer=self.output_serializers.get(k)
+                ) for k, v in namespace.items()}
         else:
             return {}
 
+    async def _serialize_namespace(self, job: Job, namespace: MutableMapping[Text, Any]) -> Text:
+        for name, value in namespace.items():
+            if name in self.input_serializers:
+                intermediate_type = self.input_serializers[name].get('type', 'name')
+                if intermediate_type == 'file':
+                    namespace[name] = dill.dumps(await self._transfer_file(job, dill.loads(value)))
+        try:
+            return await self._serialize_to_remote_file(job, namespace)
+        except BaseException:
+            for k, v in namespace.items():
+                if not dill.pickles(v, safe=True):
+                    raise WorkflowExecutionException("Name {name} is not serializable".format(name=k))
+            raise
+
     async def _serialize_to_remote_file(self, job: Job, element: Any) -> Text:
         with NamedTemporaryFile() as f:
             dill.dump(element, f, recurse=True)
@@ -95,38 +122,48 @@ class JupyterCommand(Command):
         executor_path = await self._transfer_file(
             job,
             os.path.join(executor.__file__))
-        # Modify code and namespaces according to inputs
-        updated_names = {}
-        for element in [t.value for t in job.inputs]:
-            if 'name' in element:
-                updated_names[element['name']] = element['dst']
-            else:
-                pass  # TODO: manipulate AST
+        # Modify code, environment and namespaces according to inputs
+        input_names = {}
+        environment = {}
+        for element in [t.value for t in job.inputs if t.value is not None]:
+            element_type = element['type']
+            if element_type == 'file':
+                if 'name' in element:
+                    input_names[element['name']] = element['dst']
+                else:
+                    pass  # TODO: manipulate AST
+            elif element_type == 'name':
+                input_names[element['name']] = element['value']
+            elif element_type == 'env':
+                environment[element['name']] = element['value']
+        # List output names to be retrieved from remote context
+        output_names = self.step.output_ports.keys()
         # Serialize AST nodes to remote resource
         code_path = await self._serialize_to_remote_file(job, self.ast_nodes)
-        # Vonfigure output fiel path
+        # Configure output fiel path
         path_processor = get_path_processor(self.step)
         output_path = path_processor.join(job.output_directory, random_name())
         # Serialize namespaces to remote resource
-        user_ns = {**self.user_ns, **updated_names}
-        user_ns_path = await self._serialize_to_remote_file(
-            job,
-            {k: v for k, v in user_ns.items() if k in self.input_names})
-        if self.user_global_ns != self.user_ns:
-            user_global_ns = {**self.user_global_ns, **updated_names}
-            user_global_ns_path = await self._serialize_to_remote_file(
-                job,
-                {k: v for k, v in user_global_ns.items() if k in self.input_names})
-        else:
-            user_global_ns_path = None
+        user_ns_path = await self._serialize_namespace(job=job, namespace=input_names)
+        # Create dictionaries of postload input serializers and predump output serializers
+        postload_input_serializers = {k: {'postload': v['postload']}
+                                      for k, v in self.input_serializers.items() if 'postload' in v}
+        predump_output_serializers = {k: {'predump': v['predump']}
+                                      for k, v in self.output_serializers.items() if 'predump' in v}
         # Parse command
         cmd = [self.interpreter, executor_path]
+        if self.interpreter == 'ipython':
+            cmd.append('--')
         if self.autoawait:
             cmd.append("--autoawait")
         cmd.extend(["--local-ns-file", user_ns_path])
-        if user_global_ns_path is not None:
-            cmd.extend(["--global-ns-file", user_global_ns_path])
-        for name in self.output_names:
+        if postload_input_serializers:
+            postload_serializers_path = await self._serialize_to_remote_file(job, postload_input_serializers)
+            cmd.extend(["--postload-input-serializers", postload_serializers_path])
+        if predump_output_serializers:
+            predump_serializers_path = await self._serialize_to_remote_file(job, predump_output_serializers)
+            cmd.extend(["--predump-output-serializers", predump_serializers_path])
+        for name in output_names:
             cmd.extend(["--output-name", name])
         cmd.extend([code_path, output_path])
         # Execute command
@@ -139,11 +176,10 @@ class JupyterCommand(Command):
                 command=' \\\n\t'.join(cmd),
             ))
             # If step is assigned to multiple resources, add the STREAMFLOW_HOSTS environment variable
-            parsed_env = self.environment
             if len(resources) > 1:
                 available_resources = await connector.get_available_resources(self.step.target.service)
                 hosts = {k: v.hostname for k, v in available_resources.items() if k in resources}
-                parsed_env['STREAMFLOW_HOSTS'] = ','.join(hosts.values())
+                environment['STREAMFLOW_HOSTS'] = ','.join(hosts.values())
             # Configure standard streams
             stdin = self.stdin
             stdout = self.stdout if self.stdout is not None else STDOUT
@@ -152,7 +188,7 @@ class JupyterCommand(Command):
             result, exit_code = await connector.run(
                 resources[0] if resources else None,
                 cmd,
-                environment=parsed_env,
+                environment=environment,
                 workdir=job.output_directory,
                 stdin=stdin,
                 stdout=stdout,
@@ -173,7 +209,7 @@ class JupyterCommand(Command):
             proc = await asyncio.create_subprocess_exec(
                 *cmd,
                 cwd=job.output_directory,
-                env=self.environment,
+                env=environment,
                 stdin=stdin,
                 stdout=stdout,
                 stderr=stderr
@@ -201,20 +237,12 @@ class JupyterCommand(Command):
                 json_output = json.load(f)
         if status == Status.COMPLETED:
             command_stdout = json_output[executor.CELL_OUTPUT]
-            if self.user_global_ns != self.user_ns:
-                self.user_ns = {
-                    **self.user_ns,
-                    **await self._deserialize_namespace(job, json_output[executor.CELL_LOCAL_NS])}
-                self.user_global_ns = {
-                    **self.user_global_ns,
-                    **(await self._deserialize_namespace(job, json_output[executor.CELL_GLOBAL_NS])
-                       if executor.CELL_GLOBAL_NS in json_output else {})}
-            else:
-                self.user_global_ns = self.user_ns = {
-                    **self.user_ns,
-                    **await self._deserialize_namespace(job, json_output[executor.CELL_LOCAL_NS])}
+            if isinstance(command_stdout, MutableSequence):  # TODO: understand why we obtain a list here
+                command_stdout = command_stdout[0]
+            user_ns = await self._deserialize_namespace(job, json_output[executor.CELL_LOCAL_NS])
         else:
             command_stdout = json_output[executor.CELL_OUTPUT]
+            user_ns = {}
         # Print the command stdout
         if command_stdout:
             print(command_stdout)
@@ -222,5 +250,4 @@ class JupyterCommand(Command):
         return JupyterCommandOutput(
             value=command_stdout,
             status=status,
-            user_ns=self.user_ns,
-            user_global_ns=self.user_global_ns)
+            user_ns=user_ns)
